package org.netkernel.module.standard.builtin;

import java.util.HashSet;
import java.util.Set;

import org.netkernel.container.IKernel;
import org.netkernel.container.ILogger;
import org.netkernel.container.ISpaceListener;
import org.netkernel.layer0.meta.ISpaceMeta;
import org.netkernel.layer0.meta.ISpaceElements;
import org.netkernel.layer0.meta.impl.MetaRepresentationImpl;
import org.netkernel.layer0.meta.impl.SpaceElementsImpl;
import org.netkernel.layer0.nkf.*;
import org.netkernel.layer0.urii.SimpleIdentifierImpl;
import org.netkernel.layer0.util.TupleList;
import org.netkernel.module.standard.StandardModule;
import org.netkernel.module.standard.endpoint.IStandardOverlay;
import org.netkernel.module.standard.endpoint.StandardEndpointImpl;
import org.netkernel.request.IRequestResponseFields;
import org.netkernel.request.IRequestScopeLevel;
import org.netkernel.request.impl.RequestResponseFieldsImpl;
import org.netkernel.request.impl.RequestScopeLevelImpl;
import org.netkernel.urii.*;
import org.netkernel.urii.impl.Version;
import org.netkernel.util.Utils;

public class Import extends StandardEndpointImpl implements ISpaceListener, IStandardOverlay
{
	private ISpaceWithIdentity mDelegateSpace;	
	private INKFResponseReadOnly mDelegateMetaResponse;
	private boolean mIsPrivate;
	private boolean mIsOptional;
	
	public Import()
	{	declareThreadSafe();
		declareVerboseUnhandledExceptions(false);
	}	
	
	public IRequestResponseFields getState()
	{	RequestResponseFieldsImpl state=new RequestResponseFieldsImpl(super.getState());
		if (mDelegateSpace!=null)
		{	state.put(IStandardOverlay.STATE_DELEGATE_SPACE, mDelegateSpace);
		}
		return state;
	}
	
	public void onCommissionEndpoint(IKernel aKernel, ISpace aSpace) throws Exception
	{	super.onCommissionEndpoint(aKernel, aSpace);
		
		aKernel.addSpaceListener(this);
		try
		{	findImport(false);
		}
		catch (Exception e)
		{	getKernel().getLogger().logRaw(ILogger.WARNING, this, e.toString());
		}
	}	
	
	private void findImport(boolean aWarn) throws INetKernelException
	{
		String uri = ((String)getParameter(ImportPrototype.PARAM_URI)).trim();
		String min = (String)getParameter(ImportPrototype.PARAM_MIN);
		String max = (String)getParameter(ImportPrototype.PARAM_MAX);
		mIsPrivate=(Boolean)getParameter(ImportPrototype.PARAM_PRIVATE);
		mIsOptional=(Boolean)getParameter(ImportPrototype.PARAM_OPTIONAL);
		
		IVersion minv=null;
		IVersion maxv=null;
		if (min!=null)
		{	minv=new Version(min.trim());
		}
		if (max!=null)
		{	maxv=new Version(max.trim());
		}
		
		Object owner=getStandardSpace().getOwningModule();
		if (owner instanceof StandardModule)
		{	mDelegateSpace=((StandardModule)owner).getSpace(new SimpleIdentifierImpl(uri));
		}
		if (mDelegateSpace==null)
		{	mDelegateSpace=getKernel().getSpace(new SimpleIdentifierImpl(uri), minv, maxv);
		}
		this.clearInitialisationProblems();
		if (mDelegateSpace==null && !mIsOptional)
		{	ILogger logger=getKernel().getLogger();
			String msg=logger.format("MSG_STD_BAD_IMPORT", uri,getSpace().toString());
			this.setInitialisationProblem(msg);
			if (aWarn)
			{	logger.logRaw(ILogger.WARNING, this, msg);
			}
		}
	}
	
	public void onDecommissionEndpoint() throws Exception
	{	IKernel k=getKernel();
		if (k!=null)
		{	k.removeSpaceListener(this);
		}
		super.onDecommissionEndpoint();
	}
	
	protected ISpaceMeta getDelegateMeta(INKFRequestContext aContext) throws NKFException
	{	
		INKFResponseReadOnly resp = mDelegateMetaResponse;
		if (mDelegateSpace==null)
		{	return null;
		}
		else if (resp==null || resp.isExpired())
		{	INKFRequest req = aContext.createRequest("");
			req.setVerb(INKFRequestReadOnly.VERB_META);
			req.setRepresentationClass(ISpaceMeta.class);
			IRequestScopeLevel rctx = RequestScopeLevelImpl.createOrphanedRootScopeLevel(mDelegateSpace,aContext.getKernelContext().getThisKernelRequest().getRequestScope());
			req.setRequestScope(rctx);
			resp=aContext.issueRequestForResponse(req);
			mDelegateMetaResponse=resp;
		}
		else
		{	aContext.getKernelContext().declareDependency(resp);
		}
		return (ISpaceMeta)resp.getRepresentation();
	}	
	
	public ISpaceElements getParts(INKFRequestContext aContext) throws NKFException
	{	ISpaceElements result=null;
		if (mDelegateSpace!=null)
		{	try
			{	ISpaceMeta delegateMeta = getDelegateMeta(aContext);
				if (delegateMeta!=null)
				{	ISpaceElements imported=delegateMeta.getElements();
					if (imported.size()>0)
					{
						Set<IIdentifier> ids=new HashSet<IIdentifier>();
						TupleList pl=new TupleList(3,imported.size());
						for (int i=0; i<imported.size(); i++)
						{	IIdentifier id=imported.getIdentifier(i);
							if (!ids.contains(id))
							{	ids.add(id);
								pl.put(id,mDelegateSpace,this);
							}
						}
						result = new SpaceElementsImpl(pl);
					}
				}
			}
			catch (NKFException e)
			{	setInitialisationProblem(Utils.throwableToString(e));
			}
		}
		if (result==null)
		{	result= SpaceElementsImpl.EMPTY;
		}
		return result;
	}

	public void onResolve(INKFResolutionContext aContext) throws NKFException
	{	if (mDelegateSpace!=null)
		{	aContext.delegateResolution(mDelegateSpace);
		}
	}
	
	public void onRequest(INKFRequestContext aContext) throws Exception
	{	//only META requests should come here
		if (org.netkernel.layer0.util.Utils.isCircular(aContext.getKernelContext().getThisKernelRequest()))
		{	aContext.createResponseFrom(new MetaRepresentationImpl("circular", null, null, null));
			return;
		}
		INKFRequest subreq = aContext.getThisRequest().getIssuableClone();
		subreq.injectRequestScope(mDelegateSpace);
		subreq.setRepresentationClass(IMetaRepresentation.class);
		IMetaRepresentation meta = (IMetaRepresentation)aContext.issueRequest(subreq);
		if (mIsPrivate)
		{	meta=PrivateFilterEndpoint.cloneAsPrivate(meta);
		}
		aContext.createResponseFrom(meta);
	}
	
	public String toString()
	{
		StringBuffer sb=new StringBuffer(64);
		sb.append("Import [");
		try
		{	String uri = (String)getParameter(ImportPrototype.PARAM_URI);
			sb.append(uri.trim());
		} catch (NKFException e)
		{ ; }
		sb.append("]");
		return sb.toString();
	}


	public void spaceChanged()
	{	try
		{	mDelegateMetaResponse=null;
			findImport(true);
		}
		catch (INetKernelException e)
		{	mDelegateSpace=null;
			getKernel().getLogger().logRaw(ILogger.WARNING, this, e.toString());
		}
	}

	@Override
	public void setDelegateSpace(ISpace aDelegateSpace) throws NKFException
	{	
	}
}